#include <fbab.h>

/*
 * Ouvre le fichier et initialise la lecture/criture
 */
FBAB * fbab_open(char * filename, char * flags)
{
    FBAB * f = NULL;

    f = malloc(sizeof(FBAB));
    if ( f ) {
	f->file = fopen(filename,flags);
	if ( f->file == NULL ) {
	    free(f);
	    return NULL;
	}
	f->tampon = 0;
	f->position = 7;
    }
    
    return f;
}

/*
 * Force l'criture du tampon dans le fichier et ferme le fichier
 */
void fbab_close(FBAB * f)
{
    if ( f->position != 7 )
	fwrite(&(f->tampon),1,1,f->file);

    fclose(f->file);
    free(f);
}

/*
 * Force l'criture du tampon dans le fichier
 */
void fbab_flush(FBAB * f)
{
    if ( f->position != 7 ) {
	fwrite(&(f->tampon),1,1,f->file);
	f->position = 7;
	f->tampon = 0;
    }
}

/*
 * crit un bit dans le fichier f.
 * Retourne 1 pour une criture russie et 0 dans le cas contraire.
 */
inline int fbab_writebit(unsigned char bit, FBAB * f)
{
    if ( bit )
	f->tampon |= (1<<f->position);

    f->position--;

    if ( f->position == -1 ) {
	if ( fwrite(& f->tampon,sizeof(char),1,f->file) != 1 ) {
	    perror("fwrite");
	    return 0;
	}
	f->tampon = 0;
	f->position = 7;
    }

    return 1;
}

/*
 * crit un vecteur ligne binaire
 */
int fbab_writevector(matrix * v, FBAB * f)
{
    assert(v->h == 1);

    int i;

    for ( i=0 ; i < v->w ; i++ )
	if ( ! fbab_writebit(v->matrix[0][i], f) )
	    return 0;

    return 1;
}

/*
 * crit 8 bits dans f
 * Retourne 1 pour une criture russie et 0 dans le cas contraire.
 */
int fbab_write8(unsigned char data, FBAB * f)
{
    int i;

    for ( i=7 ; i != -1 ; i-- )
	if ( ! fbab_writebit(data & (1<<i), f) )
	    return 0;
    
    return 1;
}

/*
 * crit un tableau binaire en binaire
 */
int fbab_writetab(unsigned char * tab, int size, FBAB * f)
{
    int i;
    for ( i=0 ; i < size ; i++ )
	if ( ! fbab_writebit(tab[i], f) )
	    return 0;
    return 1;
}

/*
 * Lit un bit dans f et le place dans bit
 * Retourne 1 si la lecture fut possible, 0 sinon.
 */
inline int fbab_readbit(unsigned char * bit, FBAB * f)
{
    if ( f->position == 7 )
	if ( fread(&f->tampon, 1, 1, f->file) != 1 ) {
	    if ( ! feof(f->file) ) {
		perror("fread");
		return 0;
	    }
	    return 0;
	}

    *bit = (f->tampon & (1<<f->position)) ? 1 : 0;
    f->position--;

    if ( f->position == -1 )
	f->position = 7;

    return 1;
}

/*
 * Lit 8 bits de f
 * Retourne 1 si la lecture fut possible, 0 sinon.
 */
int fbab_read8(unsigned char * data, FBAB * f)
{
    int i;
    unsigned char bit;

    *data = 0;
    for ( i=7 ; i >= 0 ; i-- ) {
	if ( ! fbab_readbit(&bit, f) )
	    return 0;
	if ( bit )
	    *data |= (1<<i);
    }

    return 1;
}

