#include "hamming.h"

/*
 * Codage
 */
int hamming(char * file, int k)
{
    if ( k < 3 ) {
	fprintf(stderr,"k doit tre suprieur  3 !\n");
	return 1;
    }

    FBAB * in = NULL, * out = NULL;
    char fileout[strlen(file)+4];
    unsigned char bit;
    matrix * g = NULL,
	   * lecture = NULL,
           * prod = NULL;
    int progress = 0, // Progression dans la lecture du bloc courant
	blocsize = (1<<k)-k-1;

    // Controle
    if ( k > 255 ) {
	fprintf(stderr,"Impossible d'appliquer l'algorithme sur une valeur "
		"si grande : %d\n",k);
	return 2;
    }

    // Ouverture des fichiers
    strcpy(fileout, file);
    strcat(fileout, ".hm");

    in = fbab_open(file, "r");
    if ( ! in ) {
	fprintf(stderr,"Impossible d'ouvrir %s\n",file);
	return 1;
    }

    out = fbab_open(fileout, "w");
    if ( ! out ) {
	fprintf(stderr,"Impossible d'ouvrir %s\n",fileout);
	return 1;
    }

    // Calcul de la matrice de codage
    g = matrix_g(k);

    // Allocation de la matrice de lecture
    lecture = matrix_new(blocsize, 1);

    // On crit la valeur k
    // en esprant qu'elle ne sera pas dtruite par le canal ...
    fbab_write8(k, out);

    // Codage/criture !
    while ( fbab_readbit(&bit, in) ) {
	lecture->matrix[0][progress] = bit;
	progress = (progress+1) % blocsize;
	if ( progress == 0 ) {
	    prod = matrix_prodLM(lecture, g);
	    fbab_writevector(prod,out);
	    matrix_free(prod);
	}
    }

    fbab_close(in);
    fbab_close(out);

    matrix_free(lecture);
    matrix_free(g);

    return 0;
}

/*
 * Lit un vecteur ligne binaire et retourne l'entier correspondant
 */
static int matrixToInt(matrix * p)
{
    int val = 0, i;

    for ( i=0 ; i < p->w ; i++ ) {
	val |= p->matrix[0][i] << (32-p->h+i+1);
    }

    return val;
}

/*
 * Corrige la matrice lue en fonction de son produit avec la matrice de controle
 */
static matrix * correct(matrix * lue, matrix * prod)
{
    int badbit = matrixToInt(prod);

    lue->matrix[badbit-1][0] = ! lue->matrix[badbit-1][0];

    return NULL;
}

/*
 * Dcodage
 */
int unhamming(char * file)
{
    FBAB * in = NULL, * out = NULL;
    char fileout[strlen(file)+5];
    matrix * h = NULL, * prod = NULL, * lecture = NULL;
    unsigned char k, bit;
    int progress = 0, blocsize;

    // Ouverture des fichiers
    strcpy(fileout, file);
    strcat(fileout, ".uhm");

    in = fbab_open(file, "r");
    if ( ! in ) {
	fprintf(stderr,"Impossible d'ouvrir %s\n",file);
	return 1;
    }

    out = fbab_open(fileout, "w");
    if ( ! out ) {
	fprintf(stderr,"Impossible d'ouvrir %s\n",fileout);
	return 1;
    }

    // Lecture de k, on espre que c'est la bonne valeur...
    fbab_read8(&k, in);

    blocsize = (1<<k)-1;

    // On calcule la matrice de controle
    h = matrix_h(k);

    // Matrice de lecture
    lecture = matrix_new(1, blocsize);

    // Dcodage
    while ( fbab_readbit(&bit, in) ) {
	lecture->matrix[progress][0] = bit;
	progress = (progress+1) % blocsize;
	// Traitement du bloc lu
	if ( progress == 0 ) {
	    prod = matrix_prodMC(h,lecture);
	    if ( matrix_null(prod) ) { // pas d'erreur
		matrix_free(prod);
		prod = matrix_databits(lecture);
		fbab_writevector(prod, out); // On crit les bits lus
		matrix_free(prod);
	    } else {
		correct(lecture,prod);
		matrix_free(prod);
		prod = matrix_databits(lecture);
		fbab_writevector(prod, out); // On crit les bits lus corrigs
		matrix_free(prod);
	    }
	}
    }

    matrix_free(lecture);
    matrix_free(h);

    fbab_close(in);
    fbab_close(out);

    return 0;
}

int main(int argc, char ** argv)
{
    switch(argc) {
	case 2:
	    return unhamming(argv[1]);
	case 3:
	    return hamming(argv[1], strtol(argv[2], NULL, 10));
	default:
	    fprintf(stderr,"Usage : %s <fichier> [k]\n",argv[0]);
	    return 1;
    }
}
