#include "huffman.h"

void libere_arbre(arbre h)
{
    if ( h != NULL ) {
	libere_arbre(h->fg);
	libere_arbre(h->fd);
	free(h);
    }
}

noeud * nouveau_noeud(int lettre, noeud * pere)
{
    noeud * n = malloc(sizeof(noeud));
    
    if ( ! n )
	return NULL;

    n->lettre = lettre;
    n->poids = 1;
    n->indice = 0;
    n->fg = n->fd = NULL;
    n->pere = pere;

    return n;
}

/*
 * Recherche un noeud d'un indice infrieur  a mais de poids gal
 */
noeud * recherche_inferieur(noeud * a, arbre h)
{
    assert(a);

    file * f = NULL;

    f = initfile();
    enfile(h,&f);
    while ( f ) {
	h = defile(&f);
	if ( a == h ) {
	    liberefile(f);
	    return NULL;
	}
	if ( a->poids == h->poids ) {
	    liberefile(f);
	    return h;
	}
	if ( h->fg )
	    enfile(h->fg, &f);
	if ( h->fd )
	    enfile(h->fd, &f);
    }

    return NULL;
}

/*
 * Retourne le noeud d'une lettre donne dans l'arbre
 */
noeud * recherche(int lettre, arbre h)
{
    noeud * tmp;

    if ( h ) {
	if ( h->lettre == lettre )
	    return h;
	if ( (tmp = recherche(lettre, h->fg)) )
	    return tmp;
	return recherche(lettre, h->fd);
    }

    return NULL;
}

int ajout_lettre(int lettre, arbre h)
{
    noeud * ajout = recherche(lettre, h), 
	  * inf = NULL,
	  * tmp = NULL;

    // lettre inconnue, on cherche NEW et le met  jour
    if ( ! ajout ) {
	ajout = recherche(-1, h);
	if ( ! ajout )
	    return 0;
	ajout->fg = nouveau_noeud(lettre, ajout);
	if ( ! ajout->fg )
	    return 0;
	ajout->fd = nouveau_noeud(-1, ajout);
	if ( ! ajout->fd )
	    return 0;
	ajout->fd->poids = 0;
	ajout->lettre = -2;

    }

    // Propagation d'information de poids
    while ( ajout != NULL ) {
	inf = recherche_inferieur(ajout, h);
	if ( inf ) { // quilibrage
	    if ( ajout->pere != inf ) {
		// On informe les pres
		if ( ajout->pere->fg == ajout )
		    ajout->pere->fg = inf;
		else
		    ajout->pere->fd = inf;

		if ( inf->pere->fg == inf )
		    inf->pere->fg = ajout;
		else
		    inf->pere->fd = ajout;

		// On change les pres
		tmp = inf->pere;
		inf->pere = ajout->pere;
		ajout->pere = tmp;
	    }
	}
	ajout->poids++;
	ajout = ajout->pere;
    }

    return 1;
}

int ecrire_lettre(int lecture, arbre h, FBAB * out)
{
    static unsigned char tab[256]; // tableau binaire
    static int taille = 0; // taille de tab
    int pos = taille;
    int ret;

    if ( h ) {
	if ( h->lettre == lecture ) {
	    if ( ! fbab_writetab(tab, taille, out) )
		return 0;
	    taille = 0;
	    return 2;
	}

	taille++;

	tab[pos] = 0; // on va  gauche
	if ( (ret = ecrire_lettre(lecture, h->fg, out)) == 2 )
	    return 2; // criture russie, propagation de l'information
	if ( ret == 0 ) // propagation d'erreur
	    return 0;

	tab[pos] = 1; //  droite
	if ( (ret = ecrire_lettre(lecture, h->fd, out)) == 2 )
	    return 2; // criture russie, propagation de l'information
	if ( ret == 0 ) // propagation d'erreur
	    return 0;

	taille--;
    }
    // On n'a pas pu coder la lettre car elle n'existe pas dans l'arbre
    return 1;
}

int huffman(char * fichier)
{
    char fileout[strlen(fichier)+3];
    arbre h = NULL;
    FILE * in = NULL;
    FBAB * out = NULL;
    int lecture;
    int ret;

    in = fopen(fichier, "r");
    if ( ! in )
	return 0;

    strcpy(fileout, fichier);
    strcat(fileout, ".h");

    out = fbab_open(fileout, "w");
    if ( ! out ) {
	fclose(in);
	return 0;
    }
    
    h = nouveau_noeud(-1, NULL); // noeud spcial NEW
    if ( ! h ) {
	fclose(in);
	fbab_close(out);
	return 0;
    }
    h->poids = 0; // poids nul pour le noeud spcial NEW

    while ( (lecture = fgetc(in)) != EOF ) {
	if ( (ret = ecrire_lettre(lecture, h, out)) != 2 ) {
	    if ( ret ) { // nouvel lettre, on code new puis la lettre
		if ( ! ecrire_lettre(-1, h, out) )
		    return 0;
		fbab_write8(lecture, out);
	    } else {
		fprintf(stderr,"Codage impossible. Abandon\n");
		return 0;
	    }
	}
	if ( ! ajout_lettre(lecture, h) ) {
	    fprintf(stderr,"Ajout dans l'arbre impossible, abandon.\n");
	    return 0;
	}
    }

    // On crit NEW pour faire la fin de fichier
    ecrire_lettre(-1, h, out);

    libere_arbre(h);
    fclose(in);
    fbab_close(out);

    return 1;
}

/*
 * Lit une lettre en fonction de l'arbre de huffman donn
 * retourne -2 en cas de fin de fichier
 */
int lire_lettre(FBAB * in, arbre h)
{    
    unsigned char bit;

    while ( h->fg != NULL && h->fd != NULL ) {
	if ( ! fbab_readbit(&bit, in) ) {
	    return -2;
	}
	if ( bit )
	    h = h->fd;
	else
	    h = h->fg;
    }

    return h->lettre;
}

int unhuffman(char * fichier)
{
    char fileout[strlen(fichier)+4];
    arbre h = NULL;
    FILE * out = NULL;
    FBAB * in = NULL;
    int lecture;
    unsigned char byte;

    in = fbab_open(fichier, "r");
    if ( ! in )
	return 0;

    strcpy(fileout, fichier);
    strcat(fileout, ".uh");

    out = fopen(fileout, "w");
    if ( ! out ) {
	fbab_close(in);
	return 0;
    }
    
    h = nouveau_noeud(-1, NULL); // noeud spcial NEW
    if ( ! h ) {
	fbab_close(in);
	fclose(out);
	return 0;
    }
    h->poids = 0; // poids nul pour le noeud spcial NEW

    while ( (lecture = lire_lettre(in,h) ) != -2 ) {
	if ( lecture == -1 ) { // NEW ?
	    if ( ! fbab_read8(&byte, in) )
		break;
	    lecture = byte;
	}
	fputc(lecture, out);
	ajout_lettre(lecture, h);
    }

    libere_arbre(h);
    fclose(out);
    fbab_close(in);

    return 1;
}

int main(int argc, char ** argv)
{
    if ( argc != 2 ) {
	fprintf(stderr,"Usage : %s <fichier>\n",argv[0]);
	return 1;
    }

    if ( strcmp(basename(argv[0]), "huffman") == 0 ) {
	return ! huffman(argv[1]);
    } else if ( strcmp(basename(argv[0]), "unhuffman") == 0 ) {
	return ! unhuffman(argv[1]);
    } else {
	fprintf(stderr,"Fonction inconnue : %s\n",argv[0]);
	return 1;
    }
}
